"""Tests for build.py."""
import pytest
import pytest_check as check

from cliconfig.routines import (
    _del_key,
    clean_pre_flat,
    flat_config,
    merge_config,
    merge_config_file,
    unflat_config,
)


def test_merge_config() -> None:
    """Test merge_config."""
    config1 = {"a.b": 1, "a": {"c": 2}}
    config2 = {"c": 3}
    config = merge_config(config1, config2, allow_new_keys=True)
    check.equal(config, {"a": {"b": 1, "c": 2}, "c": 3})
    with pytest.raises(ValueError, match="New parameter found 'c'.*"):
        merge_config(config1, config2, allow_new_keys=False)
    config1 = {"a.b": 1, "a": {"b": 1, "c": 2}}
    with pytest.raises(
        ValueError,
        match="Duplicated key found.*You may consider calling `clean_pre_flat`.*",
    ):
        merge_config(config1, config2, allow_new_keys=True)


def test_merge_config_file() -> None:
    """Test merge_config."""
    config1 = {
        "param1": 1,
        "param2": 2,
        "letters": {
            "letter1": "a",
            "letter2": "b",
        },
    }
    config2 = {
        "param3": 3,
        "letters": {
            "letter3": "c",
            "letter4": "d",
        },
    }
    expected_config = {
        "param1": 1,
        "param2": 2,
        "param3": 3,
        "letters": {
            "letter1": "a",
            "letter2": "b",
            "letter3": "c",
            "letter4": "d",
        },
    }

    config = merge_config_file(
        "tests/configs/default1.yaml",
        "tests/configs/default2.yaml",
        allow_new_keys=True,
    )
    check.equal(config, expected_config)
    config = merge_config_file(config1, config2, allow_new_keys=True)
    check.equal(config, expected_config)
    config = merge_config_file(
        config1, "tests/configs/default2.yaml", allow_new_keys=True
    )
    check.equal(config, expected_config)
    with pytest.raises(ValueError, match="New parameter found 'param3'.*"):
        merge_config_file(
            "tests/configs/default1.yaml",
            "tests/configs/default2.yaml",
            allow_new_keys=False,
        )


def test_flat_config() -> None:
    """Test flat_config."""
    check.equal(
        flat_config({"a.b": 1, "a": {"c": 2}, "d": 3}),
        {"a.b": 1, "a.c": 2, "d": 3},
    )
    check.equal(
        flat_config({"a.b": 1, "a": {"c": {}}, "a.c": 3}),
        {"a.b": 1, "a.c": 3},
    )
    with pytest.raises(ValueError, match="duplicated key 'a.b'"):
        flat_config({"a.b": 1, "a": {"b": 1}})


def test_unflat_config() -> None:
    """Test unflat_config."""
    check.equal(
        unflat_config({"a.b": 1, "a.c": 2, "c": 3}),
        {"a": {"b": 1, "c": 2}, "c": 3},
    )
    with pytest.raises(ValueError, match="The config must be flatten.*"):
        unflat_config({"a.b": 1, "a": {"c": 2}})


def test_clean_pre_flat() -> None:
    """Test clean_pre_flat."""
    check.equal(
        clean_pre_flat({"a.b": 1, "a": {"b": 2}, "c": 3}, priority="flat"),
        {"a.b": 1, "c": 3},
    )
    check.equal(
        clean_pre_flat({"a.b": 1, "a": {"b": 2}, "c": 3}, priority="unflat"),
        {"a": {"b": 2}, "c": 3},
    )
    with pytest.raises(ValueError, match="duplicated key 'a.b'"):
        clean_pre_flat({"a.b": 1, "a": {"b": 2}, "c": 3}, priority="error")
    with pytest.raises(
        ValueError,
        match=(
            "priority argument must be one of 'flat', 'unflat' or 'error' but "
            "found 'UNKNOWN'"
        ),
    ):
        clean_pre_flat({"a.b": 1, "a": {"b": 2}, "c": 3}, priority="UNKNOWN")


def test_del_key() -> None:
    """Test _del_key."""
    check.equal(
        _del_key({"a": {"b": {"c": 1}, "d": 2}, "a.e": 3, "a.b.c": 4}, "a.b.c"),
        {"a": {"d": 2}, "a.e": 3},
    )
    check.equal(
        _del_key(
            {"a": {"b": {"c": 1}, "d": 2}, "a.e": 3, "a.b.c": 4},
            "a.b.c",
            keep_flat=True,
        ),
        {"a": {"d": 2}, "a.e": 3, "a.b.c": 4},
    )
    check.equal(
        _del_key(
            {"a": {"b": {"c": 1}, "d": 2}, "a.e": 3, "a.b.c": 4},
            "a.b.c",
            keep_unflat=True,
        ),
        {"a": {"b": {"c": 1}, "d": 2}, "a.e": 3},
    )
    with pytest.raises(ValueError, match="Key 'a.b.z' not found in config."):
        _del_key({"a": {"b": {"c": 1}, "d": 2}, "a.e": 3, "a.b.c": 4}, "a.b.z")
    with pytest.raises(ValueError, match="Key 'a.z.c' not found in config."):
        _del_key({"a": {"b": {"c": 1}, "d": 2}, "a.e": 3, "a.b.c": 4}, "a.z.c")



"""CLI Config.

Copyright (C) 2023  Valentin Goldité

    This program is free software: you can redistribute it and/or modify
    it under the terms of the MIT License.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    This project is free to use for COMMERCIAL USE, MODIFICATION,
    DISTRIBUTION and PRIVATE USE as long as the original license is
    include as well as this copy right notice.
"""
# pylint: disable=wrong-import-position
from cliconfig._version import __version__, __version_tuple__
from cliconfig.build_config import make_config
from cliconfig.cli_parser import parse_cli
from cliconfig.routines import (
    clean_pre_flat,
    flat_config,
    merge_config,
    merge_config_file,
    unflat_config,
)
from cliconfig.save_load import load_config, save_config
from cliconfig.show import show_config

__all__ = [
    "__version__",
    "__version_tuple__",
    "clean_pre_flat",
    "flat_config",
    "load_config",
    "make_config",
    "merge_config",
    "merge_config_file",
    "parse_cli",
    "save_config",
    "show_config",
    "unflat_config",
]



"""Tests for save_load.py."""
import os
import shutil

import pytest
import pytest_check as check

from cliconfig.save_load import load_config, save_config


def test_save_load_config() -> None:
    """Test save_config and load_config."""
    config1 = {"a": 1, "b": {"c": 2}, "d": [2, 3.0], "e": [{"f": 4}]}
    save_config(config1, "tests/tmp/config.yaml")
    check.is_true(os.path.isfile("tests/tmp/config.yaml"))
    config2 = load_config("tests/tmp/config.yaml")
    check.equal(config1, config2)

    # With default configs
    config = {"param1": 3, "param2": 4, "letters": {"letter1": "e"}}
    save_config(config, "tests/tmp/config2.yaml")
    config = load_config(
        "tests/tmp/config2.yaml",
        default_config_paths=[
            "tests/configs/default1.yaml",
            "tests/configs/default2.yaml",
        ],
        allow_new_keys=False,
    )
    expected_config = {
        "param1": 3,
        "param2": 4,
        "param3": 3,
        "letters": {
            "letter1": "e",
            "letter2": "b",
            "letter3": "c",
            "letter4": "d",
        },
    }
    check.equal(config, expected_config)
    # Additional keys when allow_new_keys=False
    config = {"param4": 0}
    save_config(config, "tests/tmp/config3.yaml")
    with pytest.raises(ValueError, match="New parameter found 'param4'.*"):
        load_config(
            "tests/tmp/config3.yaml",
            default_config_paths=[
                "tests/configs/default1.yaml",
                "tests/configs/default2.yaml",
            ],
            allow_new_keys=False,
        )

    shutil.rmtree("tests/tmp")



"""Parser for CLI commands."""
from typing import Any, Dict, List, Tuple

import yaml


def parse_cli(sys_argv: List[str]) -> Tuple[List[str], Dict[str, Any]]:
    """Parser for CLI commands.

    Parameters
    ----------
    sys_argv : List[str]
        List of arguments from sys.argv.

    Returns
    -------
    config_paths : List[str]
        List of paths to config files to merge.
    config_cli_params : Dict[str, Any]
        Dictionary of parameters from CLI.
    """
    config_cli_params: Dict[str, Any] = {}
    config_paths: List[str] = []
    i = 0
    while i < len(sys_argv):
        elem = sys_argv[i]
        if elem == "--config":
            if config_paths:
                raise ValueError(
                    "Only one '--config ' argument is allowed in CLI (used for "
                    "config merging). If you have a subconfig called 'config', "
                    "use '--config=<val>' (with an '=')."
                )
            config_paths = sys_argv[i + 1].split(",")
            i += 2

        elif elem.startswith("--"):
            if "=" in elem:
                key, value_str = elem.split("=", maxsplit=1)
                i += 1
            else:
                key, value_str = elem, sys_argv[i + 1]
                i += 2
            key = key[2:]
            value = yaml.safe_load(value_str)
            config_cli_params[key] = value
        else:
            i += 1
    return config_paths, config_cli_params



"""Functions to build config dict."""

import sys
from typing import Any, Dict

from cliconfig.cli_parser import parse_cli
from cliconfig.routines import merge_config, merge_config_file


def make_config(
    *default_config_paths: str, allow_new_keys: bool = False
) -> Dict[str, Any]:
    r"""Make a config from default configs and CLI arguments.

    It uses the CLI Config routines to parse the CLI arguments and merge
    them in a simple way.

    Parameters
    ----------
    default_config_paths : Tuple[str]
        Paths to default configs. They are merged in order and new keys
        are allowed.
    allow_new_keys : bool, optional
        If True, new keys in CLI are allowed in merged config. Otherwise,
        it raises an error. By default False.

    Raises
    ------
    ValueError
        If allow_new_keys is False and CLI has new keys that are not
        in default configs.


    Returns
    -------
    config : Dict[str, Any]
        The merged config.

    Examples
    --------
    ::

        # main.py
        config = make_config('data.yaml', 'model.yaml', 'train.yaml')

    .. code-block:: text

        python main.py -- config bestmodel.yaml,mydata.yaml \
            --architecture.layers.hidden_dim 64

    """
    config: Dict[str, Any] = {}
    config_paths, config_cli_params = parse_cli(sys.argv)
    print(
        f"[CONFIG] Merge {len(default_config_paths)} default configs, "
        f"{len(config_paths)} additional configs and "
        f"{len(config_cli_params)} CLI parameter(s)."
    )
    for default_config_path in default_config_paths:
        config = merge_config_file(config, default_config_path)

    for config_path in config_paths:
        config = merge_config_file(config, config_path, allow_new_keys=allow_new_keys)
    config = merge_config(config, config_cli_params, allow_new_keys=allow_new_keys)
    return config



"""Saving and loading functions of configuration files."""
import os
from typing import Any, Dict, List, Optional

import yaml

from cliconfig.routines import merge_config_file


def save_config(config: Dict[str, Any], path: str) -> None:
    """Save config to a file.

    Parameters
    ----------
    config : Dict[str, Any]
        The configuration to save.
    path : str
        The path to the file to save the configuration.
    """
    dir_path = os.path.dirname(path)
    os.makedirs(dir_path, exist_ok=True)
    with open(path, "w", encoding="utf-8") as cfg_file:
        yaml.dump(config, cfg_file, default_flow_style=False)


def load_config(
    path: str,
    default_config_paths: Optional[List[str]] = None,
    *,
    allow_new_keys: bool = True,
) -> Dict[str, Any]:
    """Load config from a file and merge into optional default configs.

    Parameters
    ----------
    path : str
        The path to the file to load the configuration.
    default_config_paths : Optional[List[str]], optional
        Paths to default configs. They are merged in order, new keys are allowed.
        Then, the loaded config is merged into the result. None for no default configs.
        By default None.
    allow_new_keys : bool, optional
        If True, the new keys in the loaded config are allowed in the merged config.
        Otherwise, it raises an error. By default True.

    Raises
    ------
    ValueError
        If allow_new_keys is False and the loaded config has new keys that are not
        in default configs.

    Returns
    -------
    config : Dict[str, Any]
        The config merged from default configs and the loaded config.
    """
    config: Dict[str, Any] = {}
    if default_config_paths:
        for config_path in default_config_paths:
            config = merge_config_file(config, config_path, allow_new_keys=True)
    config = merge_config_file(config, path, allow_new_keys=allow_new_keys)
    return config



"""Tests for build_config.py."""
import sys

import pytest
import pytest_check as check

from cliconfig.build_config import make_config


def test_make_config(capsys: pytest.CaptureFixture) -> None:
    """Test make_config."""
    sys_argv = sys.argv.copy()
    sys.argv = [
        "tests/test_make_config.py.py",
        "--config",
        "tests/configs/config1.yaml,tests/configs/config2.yaml",
        "--param2",
        "6",
    ]
    capsys.readouterr()  # Clear stdout and stderr
    config = make_config("tests/configs/default1.yaml", "tests/configs/default2.yaml")
    captured = capsys.readouterr()
    out = captured.out
    expected_config = {
        "param1": 4,
        "param2": 6,
        "param3": 3,
        "letters": {
            "letter1": "f",
            "letter2": "e",
            "letter3": "c",
            "letter4": "d",
        },
    }
    expected_out = (
        "[CONFIG] Merge 2 default configs, "
        "2 additional configs and 1 CLI parameter(s).\n"
    )
    check.equal(config, expected_config)
    check.equal(out, expected_out)

    # No default configs
    config = make_config(allow_new_keys=True)
    expected_config = {
        "param1": 4,
        "param2": 6,
        "letters": {
            "letter1": "f",
            "letter2": "e",
        },
    }
    check.equal(config, expected_config)

    # No additional configs
    sys.argv = [
        "tests/test_make_config.py.py",
    ]
    config = make_config("tests/configs/default1.yaml", "tests/configs/default2.yaml")
    expected_config = {
        "param1": 1,
        "param2": 2,
        "param3": 3,
        "letters": {
            "letter1": "a",
            "letter2": "b",
            "letter3": "c",
            "letter4": "d",
        },
    }
    check.equal(config, expected_config)

    sys.argv = sys_argv.copy()



"""Routines to manipulate configuration as mixed nested and flat dictionaries."""

from typing import Any, Dict, Union

import yaml
from flatten_dict import flatten, unflatten


def merge_config(
    config1: Dict[str, Any],
    config2: Dict[str, Any],
    *,
    allow_new_keys: bool = True,
) -> Dict[str, Any]:
    """Flatten, merge config2 into config1, then unflatten merged config.

    Work even if config1 and config2 have some flat keys (like `{'a.b': ...}`).

    Parameters
    ----------
    config1 : Dict[str, Any]
        The first configuration.
    config2 : Dict[str, Any]
        The second configuration to merge into config1.
    allow_new_keys : bool, optional
        If True, new keys (that are not in config1) are allowed in config2.
        By default True.

    Raises
    ------
    ValueError
        If allow_new_keys is False and config2 has new keys that are not in config1.
    ValueError
        If there are conflicting keys when flatten one of the config.
        See examples. You may consider calling :func:`clean_pre_flat` on the input
        configs in that case.

    Returns
    -------
    config : Dict[str, Any]
        The merged nested/unflatten config (no flat keys).

    Examples
    --------
    ::

        >>> merge_config({'a.b': 1, 'a': {'c': 2}},  {'c': 3}, allow_new_keys=True)
        {'a': {'b': 1, 'c': 2}, 'c': 3}
        >>> merge_config({'a.b': 1, 'a': {'c': 2}},  {'c': 3}, allow_new_keys=False)
        ValueError: New parameter found 'c' that is not in the original config.
        >>> merge_config({'a.b': 1, 'a': {'b': 1}},  {'c': 3}, allow_new_keys=True)
        ValueError: duplicated key 'a.b'.
        The above exception was the direct cause of the following exception:
        ValueError: You may consider calling `clean_pre_flat` on config1 before merging.
    """
    # Flatten configs
    flatten_configs = []
    for i, config in enumerate([config1, config2]):
        try:
            flatten_config = flat_config(config)
        except ValueError as exc:
            raise ValueError(
                f"Duplicated key found in config {i + 1} when flattening. "
                f"You may consider calling `clean_pre_flat` before merging."
            ) from exc
        flatten_configs.append(flatten_config)
    config1_flat, config2_flat = flatten_configs[0], flatten_configs[1]

    if not allow_new_keys:
        # Check that there are no new keys in config2
        for key in config2_flat:
            if key not in config1_flat.keys():
                raise ValueError(
                    f"New parameter found '{key}' in that is not in the original "
                    "config."
                )
    # Merge flat configs
    flatten_config = {**config1_flat, **config2_flat}
    # Unflatten config
    config = unflat_config(flatten_config)
    return config


def merge_config_file(
    config_or_path1: Union[str, Dict[str, Any]],
    config_or_path2: Union[str, Dict[str, Any]],
    *,
    allow_new_keys: bool = True,
) -> Dict[str, Any]:
    """Merge two configs or their paths into one (config2 into config1).

    Similar to :func:`cliconfig.routines.merge_config` but allow passing
    the paths of config as inputs.

    Parameters
    ----------
    config_or_path1 : Union[str, Dict[str, Any]]
        The first configuration or its path.
    config_or_path2 : Union[str, Dict[str, Any]]
        The second configuration or its path to merge into first configuration.
    allow_new_keys : bool, optional
        If True, new keys (that are not in config1) are allowed in config2.
        By default True.

    Returns
    -------
    config : Dict[str, Any]
        The merged nested config (no flat keys).
    """
    configs = []
    for config_or_path in [config_or_path1, config_or_path2]:
        if isinstance(config_or_path, str):
            path = config_or_path
            with open(path, "r", encoding="utf-8") as cfg_file:
                config = yaml.safe_load(cfg_file)
        else:
            config = config_or_path
        configs.append(config)
    config1, config2 = configs[0], configs[1]
    config = merge_config(
        config1,
        config2,
        allow_new_keys=allow_new_keys,
    )
    return config


def flat_config(config: Dict[str, Any]) -> Dict[str, Any]:
    """Flatten config then return it.

    Work even if config has some flat keys as long as there are no
    conflicts (see examples section).

    Parameters
    ----------
    config : Dict[str, Any]
        The configuration to flatten.

    Raises
    ------
    ValueError
        If config has some conflicting keys (like `{'a.b': ..., 'a': {'b': ...}}`).

    Returns
    -------
    config : Dict[str, Any]
        The flattened config.

    Note
    ----
        Nested empty dict are ignored even if they are conflicting (see last example).

    Examples
    --------
    ::

        >>> flat_config({'a.b': 1, 'a': {'c': 2}, 'd': 3})
        {'a.b': 1, 'a.c': 2, 'd': 3}
        >>> flat_config({'a.b': 1, 'a': {'b': 1}})
        ValueError: duplicated key 'a.b'
        >>> flat_config({'a.b': 1, 'a': {'c': {}}, 'a.c': 3})
        {'a.b': 1, 'a.c': 3}

    """
    return flatten(config, reducer="dot")


def unflat_config(flatten_config: Dict[str, Any]) -> Dict[str, Any]:
    """Flatten a flat config then return it.

    Parameters
    ----------
    flatten_config : Dict[str, Any]
        The configuration to flatten. Must be flat.

    Raises
    ------
    ValueError
        If flatten_config is not flat and then found conflicts.

    Returns
    -------
    config : Dict[str, Any]
        The flattened config.

    Examples
    --------
    ::

        >>> unflat_config({'a.b': 1, 'a.c': 2, 'c': 3})
        {'a': {'b': 1, 'c': 2}, 'c': 3}
        >>> unflat_config({'a.b': 1, 'a': {'c': 2}})
        ValueError: duplicated key 'a'
        The config must be flatten before calling unflat_config function.
    """
    try:
        config = unflatten(flatten_config, splitter="dot")
    except ValueError as exc:
        raise ValueError(
            "The config must be flatten before calling unflat_config function."
        ) from exc
    return config


def clean_pre_flat(config: Dict[str, Any], priority: str) -> Dict[str, Any]:
    """Remove keys in config that may cause conflicts when flattening.

    Parameters
    ----------
    config: Dict[str, Any]
        The configuration to clean.
    priority: str
        One of 'flat' or 'unflat', 'error'.
        If 'flat', keys with dots at the root like `{'a.b': ...}` (flat keys) have
        priority over unflat keys like `{'a': {'b': ...}}` when there are conflicts.
        If 'unflat', unflat keys have priority over flat keys when there are conflicts.
        If 'error', raise an error when there are conflicts.

    Raises
    ------
    ValueError
        If priority is not one of 'flat', 'unflat' or 'nothing'.

    Returns
    -------
    config: Dict[str, Any]
        The cleansed configuration.

    Examples
    --------
    ::

        >>> clean_pre_flat({'a.b': 1, 'a': {'b': 2}, 'c': 3}, priority='flat')
        {'a.b': 1, 'c': 3}
        >>> clean_pre_flat({'a.b': 1, 'a': {'b': 2}, 'c': 3}, priority='unflat')
        {'a': {'b': 2}, 'c': 3}
        >>> clean_pre_flat({'a.b': 1, 'a': {'b': 2}, 'c': 3}, priority='error')
        ValueError: duplicated key 'a.b'
    """
    if priority in ("flat", "unflat"):
        # Check that there are no conflicts
        flat_part = dict(filter(lambda items: "." in items[0], config.items()))
        unflat_part = dict(filter(lambda items: "." not in items[0], config.items()))
        unflat_part_flat = flat_config(unflat_part)
        if priority == "unflat":
            for key in flat_part:
                if key in unflat_part_flat:
                    _del_key(config, key, keep_flat=False, keep_unflat=True)
        else:
            for key in unflat_part_flat:
                if key in flat_part:
                    _del_key(config, key, keep_flat=True, keep_unflat=False)
    elif priority == "error":
        flat_config(config)  # Will raise an error if there are conflicts
    else:
        raise ValueError(
            "priority argument must be one of 'flat', 'unflat' or 'error' but "
            f"found '{priority}'."
        )
    return config


def _del_key(
    config: Dict[str, Any],
    flat_key: str,
    *,
    keep_flat: bool = False,
    keep_unflat: bool = False,
) -> Dict[str, Any]:
    """Remove a value in config dict corresponding to a flat key (e.g. 'a.b.c).

    Parameters
    ----------
    config: Dict[str, Any]
        The configuration to clean. There could be a mix of flat and unflat keys.
    flat_key: str
        The flat key to remove. E.g. 'a.b.c'.
    keep_flat: bool, optional
        If True, keep the flat key in the config dict. By default False.
    keep_unflat: bool, optional
        If True, keep the unflat key in the config dict. By default False.

    Raises
    ------
    ValueError
        If the key is not found in the config dict.

    Returns
    -------
    config: Dict[str, Any]
        The new configuration.

    Examples
    --------
    ::

        >>> _del_key({'a': {'b': {'c': 1}, 'd': 2}, 'a.e': 3, 'a.b.c': 4}, 'a.b.c')
        {'a': {'d': 2}, 'a.e': 3}
        >>> _del_key({'a': {'b': {'c': 1}, 'd': 2}, 'a.e': 3, 'a.b.c': 4}, 'a.b.c',
                    keep_flat=True)
        {'a': {'d': 2}, 'a.e': 3, 'a.b.c': 4}
        >>> _del_key({'a': {'b': {'c': 1}, 'd': 2}, 'a.e': 3, 'a.b.c': 4}, 'a.b.c',
                    keep_unflat=True)
        {'a': {'b': {'c': 1}, 'd': 2}, 'a.e': 3}
        >>> _del_key({'a': {'b': {'c': 1}, 'd': 2}, 'a.e': 3, 'a.b.c': 4}, 'a.b.z')
        ValueError: Key 'a.b.z' not found in config.
        >>> _del_key({'a': {'b': {'c': 1}, 'd': 2}, 'a.e': 3, 'a.b.c': 4}, 'a.z.c')
        ValueError: Key 'a.z.c' not found in config.

    """
    found_key = False
    if not keep_flat and flat_key in config:
        # Remove flat_key if it exists at the root
        found_key = True
        del config[flat_key]

    def recursive_del_key(config: Dict[str, Any], key: str, *, found_key: bool) -> bool:
        first_key, *other_keys = key.split(".", 1)
        if other_keys:
            if first_key in config:
                # Delete key in sub-dict
                new_key = recursive_del_key(
                    config[first_key],
                    ".".join(other_keys),
                    found_key=found_key,
                )
                found_key = found_key or new_key
                # Remove if empty
                if config[first_key] == {}:
                    del config[first_key]
            else:
                # No key found, return input found_key
                return found_key
        else:
            if first_key in config:
                found_key = True
                del config[first_key]
            else:
                # No key found, return input found_key
                return found_key
        return found_key

    if not keep_unflat:
        # Remove flat_key if it exists in a nested dict
        found_key = recursive_del_key(config, flat_key, found_key=found_key)
    # Raise error if key not found
    if not found_key:
        raise ValueError(f"Key '{flat_key}' not found in config.")
    return config
